#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#define MY_EOF_CHAR	256

bool putback_char_p = 0;
int putback_char;

int prev_char_num;
int prev_line_num;
int char_num = 1;
int line_num = 1;
int s_char_num;
int s_line_num;

int next_char (ifstream &fstream) {
  int ch;
  if (putback_char_p) {
    putback_char_p = 0;
    ch = putback_char;
  }
  else
    ch = fstream.get();

  if (ch == EOF) ch = MY_EOF_CHAR;

  prev_char_num = char_num;
  prev_line_num = line_num;

  if (ch == '\n') {
    line_num++;
    char_num = 1;
  }
  else
    char_num++;

  return ch;
}

bool put_back_char (int ch) {
  if (putback_char_p) {
    cerr << "Error!  Cannot put back twice!" << endl;
    return 0;
  }
  else {
    char_num = prev_char_num;
    line_num = prev_line_num;
    putback_char_p = 1;
    putback_char = ch;
    return 1;
  }
}

#define No_State	0
#define Start_State	1
#define ID_State	2
#define TMP_State	3
#define Final_State	4

#define T_NONE		0
#define T_EOF           1
#define T_PLUS          11
#define T_TIMES         12
#define T_EQUALS        13
#define T_ASSIGN        21
#define T_IDENT         31

#define No_Action		0
#define PutBack_1_Action	1
#define Report_Error_Action	2

int yytext_loc = 0;
char yytext[1000];

class TableEntry {
  public:
    int result_action;
    int result_state;
    int return_token;
};
  
TableEntry scan_table[5][MY_EOF_CHAR+1];

void initialize_scan_table () {

  for (int ch = 0; ch <= MY_EOF_CHAR; ch++) {
    scan_table[Start_State][ch].result_action = Report_Error_Action;
    scan_table[Start_State][ch].result_state = Start_State;
    scan_table[Start_State][ch].return_token = T_NONE;
    scan_table[ID_State][ch].result_action = PutBack_1_Action;
    scan_table[ID_State][ch].result_state = Final_State;
    scan_table[ID_State][ch].return_token = T_IDENT;
    scan_table[TMP_State][ch].result_action = PutBack_1_Action;
    scan_table[TMP_State][ch].result_state = Final_State;
    scan_table[TMP_State][ch].return_token = T_ASSIGN;
  }
 
  for (int ch = 'A'; ch <= 'Z'; ch++) {
    scan_table[Start_State][ch].result_action = No_Action;
    scan_table[Start_State][ch].result_state = ID_State;
    scan_table[ID_State][ch].result_action = No_Action;
    scan_table[ID_State][ch].result_state = ID_State;
  }
  for (int ch = 'a'; ch <= 'z'; ch++) {
    scan_table[Start_State][ch].result_action = No_Action;
    scan_table[Start_State][ch].result_state = ID_State;
    scan_table[ID_State][ch].result_action = No_Action;
    scan_table[ID_State][ch].result_state = ID_State;
  }

  for (int ch = '0'; ch <= '9'; ch++) {
    scan_table[ID_State][ch].result_action = No_Action;
    scan_table[ID_State][ch].result_state = ID_State;
  }

  scan_table[Start_State]['+'].result_action = No_Action;
  scan_table[Start_State]['+'].result_state = Final_State;
  scan_table[Start_State]['+'].return_token = T_PLUS;

  scan_table[Start_State]['*'].result_action = No_Action;
  scan_table[Start_State]['*'].result_state = Final_State;
  scan_table[Start_State]['*'].return_token = T_TIMES;

  scan_table[Start_State]['='].result_action = No_Action;
  scan_table[Start_State]['='].result_state = TMP_State;

  scan_table[TMP_State]['='].result_action = No_Action;
  scan_table[TMP_State]['='].result_state = Final_State;
  scan_table[TMP_State]['='].return_token = T_EQUALS;
 
  scan_table[Start_State][MY_EOF_CHAR].result_action = No_Action;
  scan_table[Start_State][MY_EOF_CHAR].result_state = Final_State;
  scan_table[Start_State][MY_EOF_CHAR].return_token = T_EOF;
 
  scan_table[Start_State][' '].result_action = No_Action;
  scan_table[Start_State][' '].result_state = Start_State;
 
  scan_table[Start_State]['\t'].result_action = No_Action;
  scan_table[Start_State]['\t'].result_state = Start_State;
 
  scan_table[Start_State]['\n'].result_action = No_Action;
  scan_table[Start_State]['\n'].result_state = Start_State;

}

ifstream yyin;

int yylex () {
  int result_token;
  int curr_state = Start_State;
  int ch;

  yytext_loc = 0;
  yytext[yytext_loc] = '\0';

  s_line_num = line_num;
  s_char_num = char_num;

  do {
    ch = next_char(yyin);
    switch (scan_table[curr_state][ch].result_action) {
      case No_Action:
        yytext[yytext_loc] = (char) ch;
        yytext_loc++;
        yytext[yytext_loc] = '\0';
        break;
      case PutBack_1_Action:
        put_back_char(ch);
        break;
      case Report_Error_Action:
        cout << "Error!  Unexpected char (" << (char) ch << ") at line " << prev_line_num << ", char " << prev_char_num << "!" << endl;
        break;
    }
    result_token = scan_table[curr_state][ch].return_token;
    curr_state = scan_table[curr_state][ch].result_state;
    if (curr_state == Start_State) {
      yytext_loc = 0;
      yytext[yytext_loc] = '\0';
      s_line_num = line_num;
      s_char_num = char_num;
    }
  } while (curr_state != Final_State);

  return result_token;
}

int main (int argc, char *argv[]) {
  if (argc != 2) {
    cout << "Useage: assign_scan <file>" << endl;
    return -1;
  }

  initialize_scan_table();

  yyin.open(argv[1]);

  if (!yyin) {
    cout << "Error!  Unable to open file " << argv[1] << endl;
    return -1;
  }

  int token_num;
  do {
    token_num = yylex();

    cout << "Token: ";
    switch (token_num) {
      case T_EOF: cout << "EOF"; break;
      case T_PLUS: cout << "+"; break;
      case T_TIMES: cout << "*"; break;
      case T_EQUALS: cout << "=="; break;
      case T_ASSIGN: cout << "="; break;
      case T_IDENT: cout << "Ident(" << yytext << ")"; break;
    }
    cout << " found starting at line " << s_line_num << " char " << s_char_num << "\n";
  } while (token_num != T_EOF);

  return 0;
}
